#!/bin/sh

# Copyright (C) 2022  Braiins Systems s.r.o.
#
# This file is part of Braiins Open-Source Initiative (BOSI).
#
# BOSI is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# Please, keep in mind that we may also license BOSI or any part thereof
# under a proprietary license. For more information on the terms and conditions
# of such proprietary license or if you have any other questions, please
# contact us at opensource@braiins.com.

OFFSET_FROM_END_MTD0_TO_MTD1=0x600000
device_offset() {
    local devices_file=$1
    local device_name=$2

    local res=0x0
    while read line; do
        if [ "$line" = "dev:    size   erasesize  name" ] ; then
            continue
        fi
        device=$(echo $line | awk '{print $1}')
        size=$(echo $line | awk '{print "0x"$2}')
        if [ "$device" = "$device_name:" ] ; then
            break
        fi
        res=$(printf "0x%X" $(($res + $size)))
    done < "$devices_file"

    # Offset of mtd1 is 0x800000, so it is mtd0_size=0x200000 + OFFSET_FROM_END_MTD0_TO_MTD1=0x600000
    if [ "$device_name" != "mtd0" ] ; then
          res=$(printf "0x%X" $(($res + $OFFSET_FROM_END_MTD0_TO_MTD1)))
    fi
    echo $res
}

# Calculate local offset form global partition offset and global desired offset
local_offset() {
    local mtd_partition_global_offset=$1
    local global_offset=$2
    printf "0x%X" $(($global_offset - $mtd_partition_global_offset))
}

stop_and_kill_service() {
    service_name=$1
    process_name=$2
    # Default max wait time is 2 seconds
    max_wait_second=${3:-2}
    /etc/init.d/$service_name stop &
    echo "Stopping service $service_name..."
    # Wait only if $process_name is running using ps and exclude the grep process.
    # By adding square brackets around the first letter, we change the grep command to
    # `grep [b]oser` for example, so this line will not match the regular expression `[b]oser`.
    while ps | grep -q "[${process_name:0:1}]${process_name:1}"; do
        sleep 1
        max_wait_second=$((max_wait_second - 1))
        # If the process is still running after `max_wait_second` seconds, kill it.
        if [ $max_wait_second -le 0 ]; then
            echo "Killing process $process_name..."
            killall -9 "$process_name" 2> /dev/null || true
            break
        fi
    done
}

# Get file size in decimal number if second parameter is 10
# otherwise return files size in hexadecimal number
file_size() {
    number_base=$2
    decimal_file_size="$(wc -c < "$1")"
    if [ "$number_base" = "10" ]; then
        echo "$decimal_file_size"
        return 0
    fi
    printf "0x%x" "$decimal_file_size"
}

set -e
set -x

ROOT="$PWD"

# Get constants
# Expect all installation files are in the same dir as install.sh script
BASEDIR=$(dirname "$0")
. "$BASEDIR/common.sh"

# Get parameter if miner check should be skipped, this is useful for testing
MINER_TEST=true
WRITE_ENV=false
while [ "$1" ]; do
    case "$1" in
    "--skip-miner-check")
        MINER_TEST=false
        ;;
    "--help")
        echo "Usage: $0 [--skip-miner-check]"
        echo "  --skip-miner-check  Skip miner check so BOS can be installed on single board"
        exit 0
        ;;
    "--write-env")
        WRITE_ENV=true
        ;;
    *)
        echo "Unknown property $1"
        exit 1
    esac
    shift
done

BOS_ENV_FILE="$BASEDIR/$BOS_ENV_FILE"
BOS_UBOOT_FILE="$BASEDIR/$BOS_UBOOT_FILE"
BOS_DTB_FILE="$BASEDIR/$BOS_DTB_FILE"
BOS_KERNEL_FILE="$BASEDIR/$BOS_KERNEL_FILE"
BOS_ROOTFS_FILE="$BASEDIR/$BOS_ROOTFS_FILE"
BOS_TOOLS_NAME="$BASEDIR/$BOS_TOOLS_NAME"
POST_INSTALL_DIR="$BASEDIR/$POST_INSTALL_DIR"
GZIP="$BASEDIR/gzip"
CPIO="$BASEDIR/cpio"

# Define installation partition
NVDATA_PARTITION_NUMBER=6
NVDATA_PARTITION="/dev/mtd$NVDATA_PARTITION_NUMBER"

# Temp directory to mount the rootfs
TMP_BOS_DIR="/tmp/bos_image"
# Define bos_platform file
BOS_PLATFORM_PATH="etc/bos_platform"
# Define bosminer file
BOSMINER_PATH="usr/bin/bosminer"

# In `/tmp` directory the installation is already unpack, so only space for
# stock `config` partition and `nand_env` is needed. Plus some margin (2MB).
# It must comfortably fit 2x content of stock `config` partition and `nand_env`
# Units of `REQUIRED_FREE_SPACE_TMP` are in KB same as `df` output
REQUIRED_FREE_SPACE_TMP=2000
SIZE_OF_CONFIG=$(df | grep '/config' | awk '{print $4}')
REQUIRED_FREE_SPACE_TMP=$((REQUIRED_FREE_SPACE_TMP + 2 * $SIZE_OF_CONFIG))

# Get local offsets
DEVICES_INFO_FILE=/proc/mtd
if [ ! -f "$DEVICES_INFO_FILE" ]; then
    echo "Error: Missing $DEVICES_INFO_FILE" >&2
    exit 1
fi

GLOBAL_NVDATA_OFFSET=$(device_offset $DEVICES_INFO_FILE "mtd$NVDATA_PARTITION_NUMBER")

# On which offset write installation files to device $NVDATA_PARTITION
# to have it on global offsets $BOS_GLOBAL_BOOT_OFFSET
LOCAL_BOS_UBOOT_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_BOS_UBOOT_OFFSET)
LOCAL_BOS_DTB_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_BOS_DTB_OFFSET)
LOCAL_BOS_KERNEL_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_BOS_KERNEL_OFFSET)
LOCAL_BOS_ROOTFS_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_BOS_ROOTFS_OFFSET)
LOCAL_ENV_TAR_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_ENV_TAR_OFFSET)
LOCAL_ENV_TAR_LENGTH_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_ENV_TAR_LENGTH_OFFSET)
LOCAL_BACKUP_NAND_ENV_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_NAND_ENV_OFFSET)
LOCAL_RECOVERY_FLAGS_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RW_FLAGS_OFFSET)
LOCAL_RO_POST_INSTALL_TAR_LENGTH_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_POST_INSTALL_TAR_LENGTH_OFFSET)
LOCAL_RO_POST_INSTALL_TAR_OFFSET=$(local_offset $GLOBAL_NVDATA_OFFSET $GLOBAL_RO_POST_INSTALL_TAR_OFFSET)

# Size of nvdata is variable because of bad block compensation. BOS layout has compensation of 9MiB.
# The min required size of nvdata is sizeof(nvdata) - sizeof(bad_block_padding) = 0xB000000
MIN_SIZE_NVDATA=0xB000000
# Define stock mtd layout min sizes and order
MTD_MIN_SIZES="bootloader=0x200000 tpl=0x800000 misc=0x200000 recovery=0x1000000 boot=0x2000000 config=0x500000 nvdata=$MIN_SIZE_NVDATA"

# Add `trap_handler` to remove temp install directory
trap_handler() {
    rc="$?"

    # Remove the temp directory only if ends with error
    # In case of end with success, the temp directory needs to be kept
    # for second stage of installation
    if [ "$rc" -eq 0 ]; then
        exit 0
    fi
    echo "Removing temp install directory"
    rm -rf "$(realpath $BASEDIR)"

    exit $rc
}

miner_check() {
    if ! $MINER_TEST; then
        echo "Skipping miner check."
        return 0;
    fi
    # Check miner
    chmod +x "$GZIP" "$CPIO"
    # Extract bosminer and platform from rootfs.img, directory will be created
    platform=$(tail -c+65 < "$BOS_ROOTFS_FILE" | $GZIP -d | $CPIO --to-stdout -i ${BOS_PLATFORM_PATH})
    tail -c+65 < "$BOS_ROOTFS_FILE" | $GZIP -d | $CPIO --directory="$TMP_BOS_DIR" -id ${BOSMINER_PATH}

    # Run bosminer with model detect
    bosminer_check=0
    ${TMP_BOS_DIR}/${BOSMINER_PATH} --platform "$platform" model --installation-compatibility-check || bosminer_check=$?
    # Remove temp directory
    rm -rf "$TMP_BOS_DIR"

    return "$bosminer_check"
}

trap "trap_handler" EXIT

# Using hidden `--write-env` parameter will exclude checks, and only write to nand_env
# This feature is being used to perform forking to background, so when SSH connection is interrupted
if ! $WRITE_ENV; then
    # 0. Check if system is suitable for BOS
    # Check the existence and functionality of all required system utilities (nanddump, ubi*, ...)
    for i in wc dd umount ubidetach flash_eraseall nandwrite nanddump sha256sum grep df awk uname cksum nohup tar ps; do
        if ! [ -x "$(command -v $i)" ]; then
          echo "Error: $i is not installed." >&2
          exit 1
        fi
    done
    if ! ubidetach -h | grep -q "\-d,"; then
        echo "Error: Unsupported ubidetach utility." >&2
        exit 1
    fi
    if ! nandwrite -h | grep -q "Usage: nandwrite \[OPTION\] MTD_DEVICE \[INPUTFILE"; then
        echo "Error: Unsupported nandwrite utility." >&2
        exit 1
    fi
    if ! nandwrite -h | grep -q "\-s addr"; then
        echo "Error: Unsupported nandwrite utility." >&2
        exit 1
    fi
    if ! nanddump -h | grep -q "Usage: nanddump \[OPTIONS\] MTD-device"; then
        echo "Error: Unsupported nanddump utility." >&2
        exit 1
    fi
    if ! nanddump -h | grep -q "\-s addr"; then
        echo "Error: Unsupported nanddump utility." >&2
        exit 1
    fi

    # Check that required files are available
    for i in $BOS_ENV_FILE $BOS_UBOOT_FILE $BOS_DTB_FILE $BOS_KERNEL_FILE $BOS_ROOTFS_FILE $BOS_TOOLS_NAME; do
        if [ ! -f $i ]; then
            echo "Error: Required file $i for installation is missing." >&2
            exit 1
        fi
    done

    # check that the control board is Antminer Amlogic
    if ! grep -q Antminer "/usr/bin/compile_time"; then
        echo "Error: This install script is only for Antminer." >&2
        exit 2
    fi
    if ! grep -q Amlogic "/proc/cpuinfo"; then
        echo "Error: This install script is only for Amlogic." >&2
        exit 2
    fi
    # Check architecture
    if ! uname -a | grep -q aarch64; then
        echo "Error: This install script is only for aarch64." >&2
        exit 2
    fi

    # Check the compatibility of NAND layout and size
    mtd_count=$(grep -c "mtd" $DEVICES_INFO_FILE)
    if [ ! "$mtd_count" -eq 7 ]; then
        echo "Error: Different count of mtd." >&2
        cat $DEVICES_INFO_FILE >&2
        exit 2
    fi

    # Check order of mtd layout
    mtd_layout_order=$(awk '(NR>1){print $4}' $DEVICES_INFO_FILE)
    mtd_expected_order=$(echo $MTD_MIN_SIZES | tr ' ' '\n' | awk -F '=' '{print "\""$1"\""}')
    if [ "$mtd_layout_order" != "$mtd_expected_order" ]; then
        echo "Error: Unsupported nand layout order." >&2
        cat $DEVICES_INFO_FILE >&2
        exit 2
    fi

    # Check min sizes of mtd layout
    # Expected list of mtd with sizes in variable $MTD_MIN_SIZES in format 'name1=hex_size name2=hex_size'
    for name_size in $MTD_MIN_SIZES; do
        mtd_name=${name_size%=*}
        min_size=${name_size#*=}
        mtd_line=$(grep "\"$mtd_name\"" $DEVICES_INFO_FILE)
        if [ "$mtd_line" = "" ]; then
            echo "Error: Missing mtd $mtd_name." >&2
            cat $DEVICES_INFO_FILE >&2
            exit 2
        fi

        mtd_size=$(echo "$mtd_line" | awk '{print "0x"$2}')
        mtd_size=$(printf "0x%X" $(($res + $mtd_size)))
        # echo "sizes: $mtd_size $min_size"
        if [ $(($mtd_size)) -lt $(($min_size)) ]; then
            echo "Error: Not enough space on $mtd_name." >&2
            cat $DEVICES_INFO_FILE >&2
            exit 1
        fi
    done

    TMP_FREE_SPACE=$(df | grep 'tmpfs.*/tmp' | awk '{print $4}')
    if [ $REQUIRED_FREE_SPACE_TMP -gt "$TMP_FREE_SPACE" ];  then
        echo "Error: Insufficient space on /tmp." >&2
        exit 1
    fi

    # Kill cgminer, bmminer and their watchdog.
    # We must do so to prevent bmminer from interfering with bosminer
    stop_and_kill_service S71monitorcg S71monitorcg
    stop_and_kill_service S70cgminer cgminer 50
    stop_and_kill_service S70cgminer bmminer
    stop_and_kill_service S71monitor-cminer S71monitor-cminer
    stop_and_kill_service S70cminer cminer 50

    # Check miner model and whether it's supported
    if ! miner_check; then
        echo "Error: Miner is not supported by BOS." >&2
        /etc/init.d/S70cgminer start
        /etc/init.d/S71monitorcg start
        /etc/init.d/S70cminer start
        exit 3
    fi

    # Stop daemons
    stop_and_kill_service S50lighttpd lighttpd
    stop_and_kill_service S52miner_act S52miner_act
    stop_and_kill_service S66monitor-recobtn monitor-recobtn
    stop_and_kill_service S01syslogd syslogd
    stop_and_kill_service S41ntp ntpd
    stop_and_kill_service S50avahi-daemon avahi-daemon

    # Check and umount and ubidetach only if mounted
    if grep -q "/dev/ubi2_0" /proc/mounts; then
        umount -l /dev/ubi2_0
        sleep 1
        ubidetach -m $NVDATA_PARTITION_NUMBER
    fi

    # 1. Flash BOS image to $NVDATA_PARTITION
    # Use this partition because it contains only non-volatile data mainly logs
    flash_eraseall $NVDATA_PARTITION
    nandwrite -p -s $LOCAL_BOS_UBOOT_OFFSET $NVDATA_PARTITION $BOS_UBOOT_FILE
    nandwrite -p -s $LOCAL_BOS_DTB_OFFSET $NVDATA_PARTITION $BOS_DTB_FILE
    nandwrite -p -s $LOCAL_BOS_KERNEL_OFFSET $NVDATA_PARTITION $BOS_KERNEL_FILE
    nandwrite -p -s $LOCAL_BOS_ROOTFS_OFFSET $NVDATA_PARTITION $BOS_ROOTFS_FILE

    # 2 Backup nand_env script and /config
    dd if=$NAND_ENV_DEVICE of=$NAND_ENV_STOCK_FILE bs=$STOCK_ENV_LENGTH count=1
    nandwrite -p -s $LOCAL_BACKUP_NAND_ENV_OFFSET $NVDATA_PARTITION $NAND_ENV_STOCK_FILE
    # Create only tar file, because stock do not have gzip
    env_path=/tmp/env
    mkdir -p "$env_path"
    cp -aR /config "$env_path/$BOS_ENV_STOCK_CONFIG"
    # Create factory configuration and add to env tar
    $BOS_TOOLS_NAME create-factory-conf -o "$env_path/bos_factory.json"
    tar -cvf $ENV_TAR -C "$env_path" .
    printf "0x%.8x\0" "$(file_size $ENV_TAR)" | nandwrite -p -s $LOCAL_ENV_TAR_LENGTH_OFFSET $NVDATA_PARTITION
    nandwrite -p -s $LOCAL_ENV_TAR_OFFSET $NVDATA_PARTITION $ENV_TAR

    # Create post_install directory
    mkdir -p "$POST_INSTALL_DIR"
    # Create stock configuration and put it in post_install directory
    $BOS_TOOLS_NAME create-stock-conf -o "$POST_INSTALL_DIR/stock_config.json"
    # Copy run.sh to post_install directory
    cp "$BASEDIR/run.sh" "$POST_INSTALL_DIR"

    # Check if post_install have ipk packages
    packages_dir="$POST_INSTALL_DIR/packages/"
    if [ -d "$packages_dir" ]; then
        # Remove all non `.ipk` files
        cd "$packages_dir"
        # Subcommand can't be in quote to delete each file
        rm -rf $(ls | grep -v '\.ipk')
        cd "$ROOT"
    fi

    # Store post-install dir in nand
    tar -cvf $POST_INSTALL_TAR -C "$BASEDIR" "$(basename "$POST_INSTALL_DIR")"

    decimal_post_install_size=$(file_size "$POST_INSTALL_TAR" 10)
    if [ "$decimal_post_install_size" -gt "$POST_INSTALL_TAR_MAX_DECIMAL_SIZE" ];  then
        echo "Error: $POST_INSTALL_TAR too big to fit in nand space." >&2
        exit 1
    fi

    printf "0x%.8x\0" "$decimal_post_install_size" | nandwrite -p -s $LOCAL_RO_POST_INSTALL_TAR_LENGTH_OFFSET $NVDATA_PARTITION
    nandwrite -p -s $LOCAL_RO_POST_INSTALL_TAR_OFFSET $NVDATA_PARTITION $POST_INSTALL_TAR

    printf "$(printf  '\\x%x' $RECOVERY_FLAG_INSTALLED)" | nandwrite -p -s $LOCAL_RECOVERY_FLAGS_OFFSET $NVDATA_PARTITION

    cd "$ROOT"
    touch "$INSTALL_LOG"
    # Change access right so user with the name `miner` can read the log in case of failed installation
    chmod 666 "$INSTALL_LOG"
    echo "Creating background task to write to $NAND_ENV_DEVICE. Log can be found in $INSTALL_LOG."
    $BOS_TOOLS_NAME daemonize /bin/sh -- -c "source $0" "$0" "--write-env" >> "$INSTALL_LOG" 2>&1 0</dev/null

    exit 0
fi

# This part of install script is run on background

# 3. Terminate dropbear
stop_and_kill_service S50dropbear dropbear

# 4. flash nand_env script
sum_env_file=$(cksum $BOS_ENV_FILE | awk '{print $1, $2}')
size_env_file=$(wc -c $BOS_ENV_FILE | awk '{print $1}')

# Write to /dev/nand_env, and check if the write is correct with check sum
dd if=$BOS_ENV_FILE bs=$size_env_file of=$NAND_ENV_DEVICE
sum_device=$(cksum $NAND_ENV_DEVICE | awk '{print $1, $2}')

echo "Removing temp install directory"
rm -rf "$(realpath $BASEDIR)"

if [ ! "$sum_device" = "$sum_env_file" ]; then
    echo "Error: Write of $NAND_ENV_DEVICE failed." >&2
    shutdown -r now
    exit 1
else
    echo "$NAND_ENV_DEVICE successfully writen."
    echo "Rebooting now."
    shutdown -r now
    exit 0
fi
